@@ -27,7 +27,7 @@ from ruamel import yaml
 from schema_salad.sourceline import SourceLine
 from schema_salad.utils import json_dumps
 
-import cwl_utils.parser_v1_1 as cwl
+import cwl_utils.parser.cwl_v1_1 as cwl
 
 
 def expand_stream_shortcuts(process: cwl.CommandLineTool) -> cwl.CommandLineTool:
@@ -42,7 +42,7 @@ def expand_stream_shortcuts(process: cwl.CommandLineTool) -> cwl.CommandLineTool
             stdout_path = process.stdout
             if not stdout_path:
                 stdout_path = str(
-                    hashlib.sha1(
+                    hashlib.sha1(  # nosec
                         json_dumps(cwl.save(process)).encode("utf-8")
                     ).hexdigest()
                 )
@@ -76,7 +76,8 @@ def clean_type_ids(
                 for field in result.items.fields:
                     field.name = field.name.split("/")[-1]
     elif isinstance(result, cwl.InputRecordSchema):
-        result.name = result.name.split("/")[-1]
+        if result.name:
+            result.name = result.name.split("/")[-1]
         if result.fields:
             for field in result.fields:
                 field.name = field.name.split("/")[-1]
@@ -85,13 +86,13 @@ def clean_type_ids(
 
 def get_expression(
     string: str, inputs: CWLObjectType, self: Optional[CWLOutputType]
-) -> Optional[Text]:
+) -> Optional[str]:
     """
     Find and return a normalized CWL expression, if any.
 
     CWL expressions in the $() form are converted to the ${} form.
     """
-    if not isinstance(string, Text):
+    if not isinstance(string, str):
         return None
     if string.strip().startswith("${"):
         return string
@@ -100,7 +101,7 @@ def get_expression(
             "cores": 0,
             "ram": 0,
             "outdir": "/root",
-            "tmpdir": "/tmp",
+            "tmpdir": "/tmp",  # nosec
             "outdirSize": 0,
             "tmpdirSize": 0,
         }
@@ -261,7 +262,7 @@ def traverse(
                         streamable=outp.streamable,
                         doc=outp.doc,
                         format=outp.format,
-                        outputSource="main/{}".format(outp_id),
+                        outputSource=f"main/{outp_id}",
                         type=outp.type,
                         extension_fields=outp.extension_fields,
                         loadingOptions=outp.loadingOptions,
@@ -413,7 +414,7 @@ def get_input_for_id(
     name = name.split("/")[-1]
 
     for inp in cast(List[cwl.CommandInputParameter], tool.inputs):
-        if inp.id.split("#")[-1].split("/")[-1] == name:
+        if inp.id and inp.id.split("#")[-1].split("/")[-1] == name:
             return inp
     if isinstance(tool, cwl.Workflow) and "/" in name:
         stepname, stem = name.split("/", 1)
@@ -698,7 +699,7 @@ def process_workflow_inputs_and_outputs(
                                 index2,
                                 raise_type=WorkflowException,
                             ).makeError(
-                                "Entry {},".format(index)
+                                f"Entry {index},"
                                 + TOPLEVEL_SF_EXPR_ERROR.format(param.id.split("#")[-1])
                             )
     return modified
@@ -764,7 +765,7 @@ def process_workflow_reqs_and_hints(
                                     envVarReq = copy.deepcopy(req)
                                     prop_reqs += (cwl.EnvVarRequirement,)
                                 newEnvDef = copy.deepcopy(envDef)
-                                newEnvDef.envValue = "$(inputs._envDef{})".format(index)
+                                newEnvDef.envValue = f"$(inputs._envDef{index})"
                                 envVarReq.envDef[index] = newEnvDef
                                 generated_envVar_reqs.append((etool_id, index))
             if req and isinstance(req, cwl.ResourceRequirement):
@@ -793,7 +794,7 @@ def process_workflow_reqs_and_hints(
                                     loadingOptions=workflow.loadingOptions,
                                 )
                                 prop_reqs += (cwl.ResourceRequirement,)
-                            setattr(resourceReq, attr, "$(inputs._{})".format(attr))
+                            setattr(resourceReq, attr, f"$(inputs._{attr})")
                             generated_res_reqs.append((etool_id, attr))
             if req and isinstance(req, cwl.InitialWorkDirRequirement):
                 if req.listing:
@@ -844,9 +845,7 @@ def process_workflow_reqs_and_hints(
                                     None,
                                     replace_etool,
                                 )
-                                iwdr.listing[
-                                    index
-                                ] = "$(inputs._iwdr_listing_{}".format(index)
+                                iwdr.listing[index] = f"$(inputs._iwdr_listing_{index}"
                                 generated_iwdr_reqs.append((etool_id, index))
                             elif isinstance(entry, cwl.Dirent):
                                 if entry.entry:
@@ -903,7 +902,7 @@ def process_workflow_reqs_and_hints(
                                                     index,
                                                     raise_type=WorkflowException,
                                                 ).makeError(
-                                                    "Entry {},".format(index)
+                                                    f"Entry {index},"
                                                     + "Invalid CWL, if 'entry' is a string, then entryName must be specified."
                                                 )
                                             expr = (
@@ -926,7 +925,7 @@ def process_workflow_reqs_and_hints(
                                         )
                                         iwdr.listing[
                                             index
-                                        ] = "$(inputs._iwdr_listing_{}".format(index)
+                                        ] = f"$(inputs._iwdr_listing_{index}"
                                         generated_iwdr_reqs.append((etool_id, index))
 
                                 elif entry.entryname:
@@ -952,7 +951,7 @@ def process_workflow_reqs_and_hints(
                                         )
                                         iwdr.listing[
                                             index
-                                        ] = "$(inputs._iwdr_listing_{}".format(index)
+                                        ] = f"$(inputs._iwdr_listing_{index}"
                                         generated_iwdr_reqs.append((etool_id, index))
                         if generated_iwdr_reqs:
                             prop_reqs += (cwl.InitialWorkDirRequirement,)
@@ -972,8 +971,8 @@ def process_workflow_reqs_and_hints(
             for entry in generated_envVar_reqs:
                 step.in_.append(
                     cwl.WorkflowStepInput(
-                        id="_envDef{}".format(entry[1]),
-                        source="{}/result".format(entry[0]),
+                        id=f"_envDef{entry[1]}",
+                        source=f"{entry[0]}/result",
                     )
                 )
 
@@ -991,8 +990,8 @@ def process_workflow_reqs_and_hints(
             for entry in generated_res_reqs:
                 step.in_.append(
                     cwl.WorkflowStepInput(
-                        id="_{}".format(entry[1]),
-                        source="{}/result".format(entry[0]),
+                        id=f"_{entry[1]}",
+                        source=f"{entry[0]}/result",
                     )
                 )
 
@@ -1011,8 +1010,8 @@ def process_workflow_reqs_and_hints(
                 for entry in generated_iwdr_reqs:
                     step.in_.append(
                         cwl.WorkflowStepInput(
-                            id="_iwdr_listing_{}".format(index),
-                            source="{}/result".format(entry[0]),
+                            id=f"_iwdr_listing_{index}",
+                            source=f"{entry[0]}/result",
                         )
                     )
             else:
@@ -1055,6 +1054,8 @@ def process_level_reqs(
     generated_res_reqs: List[Tuple[str, str]] = []
     generated_iwdr_reqs: List[Tuple[str, Union[int, str], Any]] = []
     generated_envVar_reqs: List[Tuple[str, Union[int, str]]] = []
+    if not step.id:
+        return False
     step_name = step.id.split("#", 1)[-1]
     for req_index, req in enumerate(process.requirements):
         if req and isinstance(req, cwl.EnvVarRequirement):
@@ -1079,7 +1080,7 @@ def process_level_reqs(
                             )
                             target_process.requirements[req_index][
                                 env_index
-                            ].envValue = "$(inputs._envDef{})".format(env_index)
+                            ].envValue = f"$(inputs._envDef{env_index})"
                             generated_envVar_reqs.append((etool_id, env_index))
         if req and isinstance(req, cwl.ResourceRequirement):
             for attr in cwl.ResourceRequirement.attrs:
@@ -1098,7 +1099,7 @@ def process_level_reqs(
                         setattr(
                             target_process.requirements[req_index],
                             attr,
-                            "$(inputs._{})".format(attr),
+                            f"$(inputs._{attr})",
                         )
                         generated_res_reqs.append((etool_id, attr))
 
@@ -1134,7 +1135,7 @@ def process_level_reqs(
                         step.in_.append(
                             cwl.WorkflowStepInput(
                                 id="_iwdr_listing",
-                                source="{}/result".format(etool_id),
+                                source=f"{etool_id}/result",
                             )
                         )
                         add_input_to_process(
@@ -1171,7 +1172,7 @@ def process_level_reqs(
                             )
                             target_process.requirements[req_index].listing[
                                 listing_index
-                            ] = "$(inputs._iwdr_listing_{}".format(listing_index)
+                            ] = f"$(inputs._iwdr_listing_{listing_index}"
                             generated_iwdr_reqs.append(
                                 (etool_id, listing_index, target_type)
                             )
@@ -1187,7 +1188,7 @@ def process_level_reqs(
                                         entryname = (
                                             entry.entryname
                                             if entryname_expr
-                                            else '"{}"'.format(entry.entryname)
+                                            else f'"{entry.entryname}"'
                                         )
                                         new_expression = (
                                             "${var result; var entryname = "
@@ -1263,22 +1264,16 @@ return result; }"""
                                         (etool_id, listing_index, "string")
                                     )
     for entry in generated_envVar_reqs:
-        name = "_envDef{}".format(entry[1])
-        step.in_.append(
-            cwl.WorkflowStepInput(id=name, source="{}/result".format(entry[0]))
-        )
+        name = f"_envDef{entry[1]}"
+        step.in_.append(cwl.WorkflowStepInput(id=name, source=f"{entry[0]}/result"))
         add_input_to_process(target_process, name, "string", process.loadingOptions)
     for entry in generated_res_reqs:
-        name = "_{}".format(entry[1])
-        step.in_.append(
-            cwl.WorkflowStepInput(id=name, source="{}/result".format(entry[0]))
-        )
+        name = f"_{entry[1]}"
+        step.in_.append(cwl.WorkflowStepInput(id=name, source=f"{entry[0]}/result"))
         add_input_to_process(target_process, name, "long", process.loadingOptions)
     for entry in generated_iwdr_reqs:
-        name = "_iwdr_listing_{}".format(entry[1])
-        step.in_.append(
-            cwl.WorkflowStepInput(id=name, source="{}/result".format(entry[0]))
-        )
+        name = f"_iwdr_listing_{entry[1]}"
+        step.in_.append(cwl.WorkflowStepInput(id=name, source=f"{entry[0]}/result"))
         add_input_to_process(target_process, name, entry[2], process.loadingOptions)
     return modified
 
@@ -1310,6 +1305,8 @@ def traverse_CommandLineTool(
     # don't modifiy clt, modify step.run
     target_clt = step.run
     inputs = empty_inputs(clt)
+    if not step.id:
+        return False
     step_id = step.id.split("#")[-1]
     if clt.arguments and not skip_command_line1:
         for index, arg in enumerate(clt.arguments):
@@ -1317,15 +1314,15 @@ def traverse_CommandLineTool(
                 expression = get_expression(arg, inputs, None)
                 if expression:
                     modified = True
-                    inp_id = "_arguments_{}".format(index)
-                    etool_id = "_expression_{}{}".format(step_id, inp_id)
+                    inp_id = f"_arguments_{index}"
+                    etool_id = f"_expression_{step_id}{inp_id}"
                     target_type = "Any"
                     target = cwl.WorkflowInputParameter(id=None, type=target_type)
                     replace_step_clt_expr_with_etool(
                         expression, etool_id, parent, target, step, replace_etool
                     )
                     target_clt.arguments[index] = cwl.CommandLineBinding(
-                        valueFrom="$(inputs.{})".format(inp_id)
+                        valueFrom=f"$(inputs.{inp_id})"
                     )
                     target_clt.inputs.append(
                         cwl.CommandInputParameter(
@@ -1335,7 +1332,7 @@ def traverse_CommandLineTool(
                     )
                     step.in_.append(
                         cwl.WorkflowStepInput(
-                            "{}/result".format(etool_id), None, inp_id, None, None
+                            f"{etool_id}/result", None, inp_id, None, None
                         )
                     )
                     remove_JSReq(target_clt, skip_command_line1)
@@ -1343,8 +1340,8 @@ def traverse_CommandLineTool(
                 expression = get_expression(arg.valueFrom, inputs, None)
                 if expression:
                     modified = True
-                    inp_id = "_arguments_{}".format(index)
-                    etool_id = "_expression_{}{}".format(step_id, inp_id)
+                    inp_id = f"_arguments_{index}"
+                    etool_id = f"_expression_{step_id}{inp_id}"
                     target_type = "Any"
                     target = cwl.WorkflowInputParameter(id=None, type=target_type)
                     replace_step_clt_expr_with_etool(
@@ -1360,9 +1357,7 @@ def traverse_CommandLineTool(
                         )
                     )
                     step.in_.append(
-                        cwl.WorkflowStepInput(
-                            id=inp_id, source="{}/result".format(etool_id)
-                        )
+                        cwl.WorkflowStepInput(id=inp_id, source=f"{etool_id}/result")
                     )
                     remove_JSReq(target_clt, skip_command_line1)
     for streamtype in "stdout", "stderr":  # add 'stdin' for v1.1 version
@@ -1371,21 +1366,19 @@ def traverse_CommandLineTool(
             expression = get_expression(stream_value, inputs, None)
             if expression:
                 modified = True
-                inp_id = "_{}".format(streamtype)
-                etool_id = "_expression_{}{}".format(step_id, inp_id)
+                inp_id = f"_{streamtype}"
+                etool_id = f"_expression_{step_id}{inp_id}"
                 target_type = "string"
                 target = cwl.WorkflowInputParameter(id=None, type=target_type)
                 replace_step_clt_expr_with_etool(
                     expression, etool_id, parent, target, step, replace_etool
                 )
-                setattr(target_clt, streamtype, "$(inputs.{})".format(inp_id))
+                setattr(target_clt, streamtype, f"$(inputs.{inp_id})")
                 target_clt.inputs.append(
                     cwl.CommandInputParameter(id=inp_id, type=target_type)
                 )
                 step.in_.append(
-                    cwl.WorkflowStepInput(
-                        id=inp_id, source="{}/result".format(etool_id)
-                    )
+                    cwl.WorkflowStepInput(id=inp_id, source=f"{etool_id}/result")
                 )
     for inp in clt.inputs:
         if not skip_command_line1 and inp.inputBinding and inp.inputBinding.valueFrom:
@@ -1395,19 +1388,17 @@ def traverse_CommandLineTool(
             if expression:
                 modified = True
                 self_id = inp.id.split("#")[-1]
-                inp_id = "_{}_valueFrom".format(self_id)
-                etool_id = "_expression_{}{}".format(step_id, inp_id)
+                inp_id = f"_{self_id}_valueFrom"
+                etool_id = f"_expression_{step_id}{inp_id}"
                 replace_step_clt_expr_with_etool(
                     expression, etool_id, parent, inp, step, replace_etool, self_id
                 )
-                inp.inputBinding.valueFrom = "$(inputs.{})".format(inp_id)
+                inp.inputBinding.valueFrom = f"$(inputs.{inp_id})"
                 target_clt.inputs.append(
                     cwl.CommandInputParameter(id=inp_id, type=inp.type)
                 )
                 step.in_.append(
-                    cwl.WorkflowStepInput(
-                        id=inp_id, source="{}/result".format(etool_id)
-                    )
+                    cwl.WorkflowStepInput(id=inp_id, source=f"{etool_id}/result")
                 )
     for outp in clt.outputs:
         if outp.outputBinding:
@@ -1416,13 +1407,13 @@ def traverse_CommandLineTool(
                 if expression:
                     modified = True
                     inp_id = "_{}_glob".format(outp.id.split("#")[-1])
-                    etool_id = "_expression_{}{}".format(step_id, inp_id)
+                    etool_id = f"_expression_{step_id}{inp_id}"
                     glob_target_type = ["string", cwl.ArraySchema("string", "array")]
                     target = cwl.WorkflowInputParameter(id=None, type=glob_target_type)
                     replace_step_clt_expr_with_etool(
                         expression, etool_id, parent, target, step, replace_etool
                     )
-                    outp.outputBinding.glob = "$(inputs.{})".format(inp_id)
+                    outp.outputBinding.glob = f"$(inputs.{inp_id})"
                     target_clt.inputs.append(
                         cwl.CommandInputParameter(
                             id=inp_id,
@@ -1430,9 +1421,7 @@ def traverse_CommandLineTool(
                         )
                     )
                     step.in_.append(
-                        cwl.WorkflowStepInput(
-                            id=inp_id, source="{}/result".format(etool_id)
-                        )
+                        cwl.WorkflowStepInput(id=inp_id, source=f"{etool_id}/result")
                     )
             if outp.outputBinding.outputEval and not skip_command_line2:
                 self: CWLOutputType = [
@@ -1441,8 +1430,8 @@ def traverse_CommandLineTool(
                         "basename": "base.name",
                         "nameroot": "base",
                         "nameext": "name",
-                        "path": "/tmp/base.name",
-                        "dirname": "/tmp",
+                        "path": "/tmp/base.name",  # nosec
+                        "dirname": "/tmp",  # nosec
                     }
                 ]
                 if outp.outputBinding.loadContents:
@@ -1451,8 +1440,8 @@ def traverse_CommandLineTool(
                 if expression:
                     modified = True
                     outp_id = outp.id.split("#")[-1]
-                    inp_id = "_{}_outputEval".format(outp_id)
-                    etool_id = "expression{}".format(inp_id)
+                    inp_id = f"_{outp_id}_outputEval"
+                    etool_id = f"expression{inp_id}"
                     sub_wf_outputs = cltool_step_outputs_to_workflow_outputs(
                         step, etool_id, outp_id
                     )
@@ -1502,7 +1491,7 @@ def traverse_CommandLineTool(
                     new_clt_step = copy.copy(
                         step
                     )  # a deepcopy would be convienant, but params2.cwl gives it problems
-                    new_clt_step.id = new_clt_step.id.split("#")[-1]
+                    new_clt_step.id = cast(str, new_clt_step.id).split("#")[-1]
                     new_clt_step.run = copy.copy(step.run)
                     new_clt_step.run.id = None
                     remove_JSReq(new_clt_step.run, skip_command_line1)
@@ -1554,13 +1543,11 @@ def traverse_CommandLineTool(
                     step.run = sub_workflow
                     rename_step_source(
                         sub_workflow,
-                        "{}/{}".format(step_id, outp_id),
-                        "{}/result".format(etool_id),
+                        f"{step_id}/{outp_id}",
+                        f"{etool_id}/result",
                     )
                     orig_step_inputs.append(
-                        cwl.WorkflowStepInput(
-                            id="self", source="{}/{}".format(step_id, outp_id)
-                        )
+                        cwl.WorkflowStepInput(id="self", source=f"{step_id}/{outp_id}")
                     )
                     if not parent.requirements:
                         parent.requirements = [cwl.SubworkflowFeatureRequirement()]
@@ -1730,7 +1717,7 @@ def cltool_inputs_to_etool_inputs(
 
 
 def cltool_step_outputs_to_workflow_outputs(
-    cltool_step: cwl.WorkflowStep, etool_step_id: Text, etool_out_id: str
+    cltool_step: cwl.WorkflowStep, etool_step_id: str, etool_out_id: str
 ) -> List[cwl.OutputParameter]:
     """
     Copy CommandLineTool outputs into the equivalent Workflow output parameters.
@@ -1739,14 +1726,16 @@ def cltool_step_outputs_to_workflow_outputs(
     they came from.
     """
     outputs = yaml.comments.CommentedSeq()
+    if not cltool_step.id:
+        raise WorkflowException(f"Missing step id from {cltool_step}.")
     default_step_id = cltool_step.id.split("#")[-1]
     if cltool_step.run.outputs:
         for clt_out in cltool_step.run.outputs:
             clt_out_id = clt_out.id.split("#")[-1].split("/")[-1]
             if clt_out_id == etool_out_id:
-                outputSource = "{}/result".format(etool_step_id)
+                outputSource = f"{etool_step_id}/result"
             else:
-                outputSource = "{}/{}".format(default_step_id, clt_out_id)
+                outputSource = f"{default_step_id}/{clt_out_id}"
             if not clt_out_id.startswith("_"):
                 outputs.append(
                     cwl.WorkflowOutputParameter(
@@ -1790,7 +1779,7 @@ def generate_etool_from_expr2(
     )
     expression = "${"
     if self_name:
-        expression += "\n  var self=inputs.{};".format(self_name)
+        expression += f"\n  var self=inputs.{self_name};"
     expression += (
         """
   return {"result": function(){"""
@@ -1838,6 +1827,8 @@ def traverse_step(
     """Process the given WorkflowStep."""
     modified = False
     inputs = empty_inputs(step, parent)
+    if not step.id:
+        return False
     step_id = step.id.split("#")[-1]
     original_process = copy.deepcopy(step.run)
     original_step_ins = copy.deepcopy(step.in_)
@@ -1940,7 +1931,7 @@ def traverse_step(
                     source_type,
                 )
                 inp.valueFrom = None
-                inp.source = "{}/result".format(etool_id)
+                inp.source = f"{etool_id}/result"
     # TODO: skip or special process for sub workflows?
     process_modified = process_level_reqs(
         original_process,
@@ -1972,6 +1963,8 @@ def workflow_step_to_WorkflowInputParameters(
     """Create WorkflowInputParametes to match the given WorkflowStep inputs."""
     params = []
     for inp in step_ins:
+        if not inp.id:
+            continue
         inp_id = inp.id.split("#")[-1].split("/")[-1]
         if inp.source and inp_id != except_in_id:
             param = copy.deepcopy(param_for_source_id(parent, sourcenames=inp.source))
@@ -2003,6 +1996,8 @@ def replace_step_valueFrom_expr_with_etool(
     ] = None,
 ) -> None:
     """Replace a WorkflowStep level 'valueFrom' expression with a sibling ExpressionTool step."""
+    if not step_inp.id:
+        raise WorkflowException(f"Missing id in {step_inp}.")
     step_inp_id = step_inp.id.split("/")[-1]
     etool_inputs = workflow_step_to_WorkflowInputParameters(
         original_step_ins, workflow, step_inp_id
@@ -2032,6 +2027,8 @@ def replace_step_valueFrom_expr_with_etool(
     if source:
         wf_step_inputs.append(cwl.WorkflowStepInput(id="self", source=step_inp.source))
     for wf_step_input in wf_step_inputs:
+        if not wf_step_input.id:
+            continue
         wf_step_input.id = wf_step_input.id.split("/")[-1]
         if wf_step_input.valueFrom:
             wf_step_input.valueFrom = None
@@ -2044,7 +2041,7 @@ def replace_step_valueFrom_expr_with_etool(
     wf_step_inputs[:] = [
         x
         for x in wf_step_inputs
-        if not (x.id.startswith("_") or x.id.endswith(step_inp_id))
+        if x.id and not (x.id.startswith("_") or x.id.endswith(step_inp_id))
     ]
     scatter = copy.deepcopy(step.scatter)
     if isinstance(scatter, str):
